\sokoban.xpl	11-Jul-2017
\3D Isometric Sokoban Game
\
\REVISIONS:
\04-May-2002: Released (Escaped?)
\31-May-2002: Added Sound Blaster sounds and other minor mods.
\11-Jul-2017: Modified for RPi and slayed the frumious segmentation fault!

def	Debug = true;		\enable error trapping and diagnostic messages

char	FloorPlan(20, 17),	\20 columns by 17 rows are maximum dimensions
	ScreenBuf(320*200),	\screen buffer: image that gets copied to screen
	ColorReg(256*3);	\copy of initial color register values for Fade

char	SpTitle(2+255*64),	\sprites: (byte: width, height, graphic image)
	SpFloor(2+32*32),	\most sprites are 32x32 pixels
	SpDock(2+32*32),
	SpWall(2+32*32),
	SpBox(2+32*32),
	SpManS(2+32*32),	\man facing in four directions
	SpManN(2+32*32),
	SpManW(2+32*32),
	SpManE(2+32*32);

def	StkMax = 30000;		\stack for saving & undoing 30000/11 = 2727 moves
char	Stack(StkMax),
	StkPtr;			\stack pointer (see Push and Pop procedures)

int	Screen,			\where Linux put our display mem. (frame buffer)
	Key,			\keystroke's ASCII value or scan code
	Level,			\level number of floor plan
	LevelHandle,		\input file handle to read SokoLevels.txt
	ManX, ManY,		\man's location in FloorPlan (tiles)
	NewX, NewY,		\possible new man location (tiles)
	BoxX, BoxY,		\possible new box location (tiles)
	RefX, RefY,		\reference tile location; used for panning
	MoveCtr;		\move counter (number of arrow keystrokes)



proc	Exit(Msg);		\Display (error) message and terminate program
char	Msg;
[SetVid($03);			\restore normal text mode
Text(0, Msg);
CrLf(0);
exit;
];	\Exit



proc	Erase(Im, C);		\Erase image Im to color C
int	Im, C;
[FillMem(Im, C, 320*200);
 FillMem(Im, 0, 320*8);		\top text line is black
];	\Erase



proc	Push(B);		\Push byte B onto the Stack
int	B;
[if StkPtr < StkMax then
	[Stack(StkPtr):= B;
	StkPtr:= StkPtr+1;
	];
];	\Push



func	Pop;			\Pop byte from Stack and return it
[if StkPtr > 0 then
	StkPtr:= StkPtr-1;
return Stack(StkPtr);
];	\Pop

\-------------------------------------------------------------------------------

proc	Fade(Down, Speed);	\Fade screen colors up or down
int	Down,			\flag: fade down to black
	Speed;			\0=slow, 1=fast, etc.
int	Inten;			\intensity


	proc	Set(I);		\Set intensity of the colors
	int	I;
	int	J;
	[WaitForVSync;
	for J:= 0 to 256-1 do	\set color registers
		SetPalette(J, ColorReg(J*3)*I/64, ColorReg(J*3+1)*I/64,
			      ColorReg(J*3+2)*I/64);
	];	\Set


[if Down then
	for Inten:= $3F downto 0 do [Set(Inten);  Inten:= Inten + Speed]
else	for Inten:=   1 to   $40 do [Set(Inten);  Inten:= Inten + Speed];
];	\Fade

\-------------------------------------------------------------------------------

proc	DrawSprite(X0, Y0, Sp);	\Draw a sprite onto ScreenBuf
int	X0, Y0;	\coordinates where upper-left corner of sprite is drawn
char	Sp;	\address of sprite data
int	X, Y, K, Y320, C, W, H;
[W:= Sp(0);	\get width and height (in pixels)
 H:= Sp(1);
K:= 2;
for Y:= Y0 to Y0+H-1 do
	[Y320:= Y*320;
	for X:= X0 to X0+W-1 do
		[C:= Sp(K);			\get pixel's color
		K:= K+1;
		if C & X>=0 & X<320 then	\background (0) is transparent
		    if Y>=8 & Y<200 then	\clip below text on top line
			ScreenBuf(X+Y320):= C;
		];
	];
];	\DrawSprite

\-------------------------------------------------------------------------------

proc	LoadFloorPlan;		\Load next floorplan level into FloorPlan array
int	I, J, Ch;
[FSet(LevelHandle, ^i);			\associate device 3 with levels.txt file
for J:= 0 to 17-1 do			\all floor plans are 20x17
    for I:= 0 to 20-1 do
	[repeat	Ch:= ChIn(3);
		if Ch = $1A\EOF\ then Exit("All levels solved!");  \exit program
	until Ch>=$20 & Ch<=$7E & Ch#^>;	\skip CR, LF, etc.
	case Ch of
	  ^":	FloorPlan(I,J):= $00;	\background
	  ^ :	FloorPlan(I,J):= $01;	\floor
	  ^!:	FloorPlan(I,J):= $02;	\loading dock
	  ^#:	FloorPlan(I,J):= $03;	\wall
	  ^$:	FloorPlan(I,J):= $11;	\box (high nibble) on floor (low nibble)
	  ^%:	FloorPlan(I,J):= $12;	\box on dock
	  ^&:  [FloorPlan(I,J):= $21;	\man on floor
		ManX:= I;  ManY:= J;	\record man's position
		\shift floorplan image if necessary to center it on the man
		RefX:= 0;  RefY:= 0;
		if ManX <  6 then RefX:= ManX - 6;
		if ManX > 13 then RefX:= ManX - 13;
		if ManY <  6 then RefY:= ManY - 6;
		if ManY > 13 then RefY:= ManY - 13;
	       ]
	other	[];			\ignore any other characters
	];
];	\LoadFloorPlan

\-------------------------------------------------------------------------------

proc	DrawFloorPlan;	\...for current level, including boxes and man
int	I, J,		\coordinates on floor (tiles)
	X, Y,		\coordinates on screen (pixels)
	N;		\level number


	func	Tile2PixX(I, J);    \Convert tile position to pixel X coordinate
	int	I, J;
	[I:= I - RefX;
	 J:= J - RefY;
	return J*16 - (19-I)*16 + 160-16;
	];	\Tile2PixX


	func	Tile2PixY(I, J);    \Convert tile position to pixel Y coordinate
	int	I, J;
	[I:= I - RefX;
	 J:= J - RefY;
	return J*8 - I*8 + 96-8;
	];	\Tile2PixY


	proc	DrawMan;	\...facing direction of last keystroke
	case Key of
	  $48:	DrawSprite(X, Y, SpManN);	\up
	  $4B:	DrawSprite(X, Y, SpManW);	\left
	  $4D:	DrawSprite(X, Y, SpManE)	\right
	other	DrawSprite(X, Y, SpManS);	\down


[	\DrawFloorPlan
Erase(ScreenBuf, 2\green grass background\);

\Draw tiles starting from the back (ala painter's algorithm)
for J:= 0 to 17-1 do				\from top to bottom...
    for I:= 19 downto 0 do			\ from right to left...
	[X:= Tile2PixX(I, J);
	 Y:= Tile2PixY(I, J);
	case FloorPlan(I, J) of
	  $00:	[];				\background
	  $01:	DrawSprite(X, Y, SpFloor);
	  $02:	DrawSprite(X, Y, SpDock);
	  $03:	DrawSprite(X, Y, SpWall);
	  $11:	[DrawSprite(X, Y, SpFloor);
		 DrawSprite(X, Y, SpBox)];
	  $12:	[DrawSprite(X, Y, SpDock);
		 DrawSprite(X, Y, SpBox)];
	  $21:	[DrawSprite(X, Y, SpFloor);  DrawMan];
	  $22:	[DrawSprite(X, Y, SpDock);   DrawMan]
	other	[if Debug then Exit("Illegal sprite code")];
	];
WaitForVSync;
CopyMem(Screen, ScreenBuf, 320*200);

\Display level name in upper-left corner
Attrib($0007);			\white on black (for 256-color mode)
Cursor(0, 0);
Text(6, "LEVEL ");
IntOut(6, Level);

\Display move counter in upper-right corner
Cursor(30, 0);
N:= MoveCtr;			\right-justify the number
if N = 0 then N:= 1;
while N < 1000 do [ChOut(6, ^ );  N:= N*10];
IntOut(6, MoveCtr);
Text(6, " MOVES");
];	\DrawFloorPlan

\-------------------------------------------------------------------------------

func	Solved;		\Return 'true' if all the docks have boxes on them
int	I, J;		\(Note: Some levels have more boxes than docks)
[for J:= 0 to 17-1 do
    for I:= 0 to 20-1 do
	[if (FloorPlan(I,J)&$0F) = $02\dock\ then
		if (FloorPlan(I,J)&$F0) # $10\box\ then return false;
	];
return true;
];	\Solved



proc	Undo;		\Undo a move
[if StkPtr > 0 then			\there's a move saved on the stack
       [RefY:= extend(Pop);		\restore panning position
	RefX:= extend(Pop);
	BoxY:= Pop;			\restore what was new location of box
	BoxX:= Pop;
	NewY:= Pop;			\restore what was new location of man
	NewX:= Pop;
	ManY:= Pop;			\restore location where man is
	ManX:= Pop;
	FloorPlan(BoxX,BoxY):= Pop;	\restore contents where box might go
	FloorPlan(NewX,NewY):= Pop;	\restore contents at new man location
	FloorPlan(ManX,ManY):= Pop;	\restore man and his background
	MoveCtr:= MoveCtr-1;		\undo move count
       ];
];	\Undo



proc	DoMove;		\The arrow keys move the man and possibly a box


	proc	MoveMan;	\Display man at New & blank out man at ManX,ManY
	[FloorPlan(NewX,NewY):= (FloorPlan(NewX,NewY) & $0F) ! $20;
	 FloorPlan(ManX,ManY):= FloorPlan(ManX,ManY) & $0F;
	ManX:= NewX;  ManY:= NewY;	\set man to new location

	if ManX-RefX <  6 then RefX:= RefX-1;	\pan display if necessary
	if ManX-RefX > 13 then RefX:= RefX+1;
	if ManY-RefY <  6 then RefY:= RefY-1;
	if ManY-RefY > 13 then RefY:= RefY+1;
	];	\MoveMan


[	\DoMove
NewX:= ManX;  NewY:= ManY;
BoxX:= ManX;  BoxY:= ManY;
case Key of	\determine possible new locations for man and box
  $48:	[NewY:= ManY-1;  if NewY-1 >  0 then BoxY:= NewY-1];	\up
  $4B:	[NewX:= ManX-1;  if NewX-1 >  0 then BoxX:= NewX-1];	\left
  $4D:	[NewX:= ManX+1;  if NewX+1 < 20 then BoxX:= NewX+1];	\right
  $50:	[NewY:= ManY+1;  if NewY+1 < 17 then BoxY:= NewY+1]	\down
other	[];

\Save state so the move can be undone
\When moving the man or box, preserve background color of floor or dock
Push(FloorPlan(ManX,ManY));		\save man and his background
Push(FloorPlan(NewX,NewY));		\save contents at new man location
Push(FloorPlan(BoxX,BoxY));		\save contents where box might go
Push(ManX);				\save location of man
Push(ManY);
Push(NewX);				\save new location of man
Push(NewY);
Push(BoxX);				\save new location of box
Push(BoxY);
Push(RefX);				\save panning position
Push(RefY);

MoveCtr:= MoveCtr+1;	\increment move counter (even if move is illegal)

if FloorPlan(NewX,NewY)=$01\floor\ ! FloorPlan(NewX,NewY)=$02\dock\ then
	MoveMan
else if (FloorPlan(NewX,NewY) & $F0) = $10\box\ then	\attempting to move box
    [if FloorPlan(BoxX,BoxY)=$01\floor\ ! FloorPlan(BoxX,BoxY)=$02\dock\ then
	[if FloorPlan(BoxX,BoxY)=$02\dock\ &
	    FloorPlan(NewX,NewY)=$11\box on floor\ then
		PlaySoundFile("SokoDocked.wav")	\box moved from floor to dock
	else	PlaySoundFile("SokoSlide.wav");
	FloorPlan(BoxX,BoxY):= (FloorPlan(BoxX,BoxY) & $0F) ! $10; \display box
	MoveMan;
	]
    else PlaySoundFile("SokoGrunt7.wav");	\two boxes or box and wall
    ]
else PlaySoundFile("SokoGrunt2.wav");		\hit wall
];	\DoMove

\-------------------------------------------------------------------------------

proc	LoadSprites;		\Load sprite images from .bmp file
int	Handle,			\file handle (descriptor)
	Width, Height,		\dimensions of entire image in .bmp file, pixels
	R, G, B,		\red, green, blue
	X, Y,			\screen coordinates (pixels)
	Y320,			\Y*320
	I, K, T;		\indexes and tempory scratch


	proc	LoadSp(X0, Y0, W, H, Sp);	\Load a sprite
	int	X0, Y0,		\coordinates in ScreenBuf to get sprite from
		W, H;		\width and height (pixels)
	char	Sp;		\sprite to load
	int	K, X, Y;
	[Sp(0):= W;
	 Sp(1):= H;
	K:= 2;
	for Y:= Y0 to Y0+H-1 do
	    for X:= X0 to X0+W-1 do
		[Sp(K):= ScreenBuf(X+Y*320);
		K:= K+1;
		];
	];	\LoadSp


[	\LoadSprites
\Read in a 256-color .bmp file
Trap(false);			\don't trap "file not found" error; we'll do it
Handle:= FOpen("SokoSprites.bmp", 0); \open file for input
if GetErr \#0\ then Exit("SokoSprites.bmp not found");
Trap(Debug);			\turn traps back on if debug mode
FSet(Handle, ^I);		\set device 3 to handle
OpenI(3);

for Y:= 0 to 17 do X:= ChIn(3);	\skip unused header info
Width:= ChIn(3) + ChIn(3)<<8;	\0..32764 (ample range)
for Y:= 0 to 2-1 do X:= ChIn(3);\skip
Height:= ChIn(3) + ChIn(3)<<8;	\0..32767 (ample range)
for Y:= 24 to 53 do X:= ChIn(3);\skip

K:= 0;
for I:= 0 to 256-1 do
	[B:= ChIn(3);
	 G:= ChIn(3);
	 R:= ChIn(3);
	 T:= ChIn(3);
	SetPalette(I, R, G, B);
	ColorReg(K):= R;	\save copy for Fade routine
	ColorReg(K+1):= G;
	ColorReg(K+2):= B;
	K:= K+3;
	];
\Load .bmp image into ScreenBuf
for Y:= Height-1 downto 0 do	\.bmp files are upside down!
	[Y320:= 320*Y;
	for X:= 0 to Width-1 do
		ScreenBuf(X+Y320):= ChIn(3);
	];
FClose(Handle);			\close handle so it can be used again

LoadSp( 0,136,255, 64, SpTitle); \grab individual sprites
LoadSp( 0,  0, 32, 32, SpFloor);
LoadSp(32,  0, 32, 32, SpDock);
LoadSp(64,  0, 32, 32, SpWall);
LoadSp(96,  0, 32, 32, SpBox);
LoadSp( 0, 32, 32, 32, SpManS);
LoadSp(32, 32, 32, 32, SpManN);
LoadSp(64, 32, 32, 32, SpManW);
LoadSp(96, 32, 32, 32, SpManE);
];	\LoadSprites

\-------------------------------------------------------------------------------

proc	GetStartingLevel;	\Display title screen and get starting level num
int	I, Ch;
[Trap(false);			\don't trap "file not found" error; we'll do it
LevelHandle:= FOpen("SokoLevels.txt", 0);	\open file for input
if GetErr \#0\ then Exit("SokoLevels.txt not found");
Trap(Debug);			\restore traps if not debug mode
FSet(LevelHandle, ^i);		\read levels with small buffer to avoid conflict
OpenI(3);

Erase(ScreenBuf, 0\black\);	\show title
DrawSprite(32, 32, SpTitle);
CopyMem(Screen, ScreenBuf, 320*200);

Attrib($20);			\bright red
Cursor(7, 17);
Text(6, "Enter starting level: __");
loop	[Attrib($0F);		\bright white
	Cursor(7+22, 17);

	Level:= 0;		\Level:= IntIn(6); doesn't do everything we want
	repeat	Ch:= ChIn(6);
		if Ch>=^0 & Ch<=^9 then
			Level:= Level*10 + Ch - ^0;
		if Ch=$1B\Esc\ ! Ch=$03\Ctrl+C\ ! Ch=0\Break\ then Exit(" ");
	until Ch<^0 ! Ch>^9;

	if Level>=1 & Level<=36 then quit;	\check for valid range
	Cursor(7, 19);
	Attrib($20);
	Text(6, "Please select 1 thru 36");
	repeat until KeyHit;	\wait for keystroke
	Cursor(7+22, 17);	\blank erroneous entry
	Text(6, "__     ");
	];
\Skip to selected level in the input file
for I:= 1 to Level do
	repeat	Ch:= ChIn(3);
	until Ch = ^>;
Fade(true, 0);			\slowly fade to black
];	\GetStartingLevel

\===============================================================================

[	\Main
TrapC(true);				\don't let Ctrl+C bypass Exit routine
SetVid($13);				\set 320x200 graphics in 256 colors
Screen:= GetFB;				\get location where Linux put our frame
Screen:= Screen(3);			\ buffer (i.e. screen image)
LoadSprites;				\load sprites and set up color registers
GetStartingLevel;

loop	[StkPtr:= 0;			\level loop
	MoveCtr:= 0;			\initialize for Undo command
	LoadFloorPlan;			\from input file
	DrawFloorPlan;
	Fade(false, 0);			\slowly fade in new level's floor plan

	loop	[Key:= ChIn(1);		\move loop
		case Key of
		  $00:	[];		\ignore scan code prefix byte
		  $03, $1B:		\Esc and Ctrl+C
			[Fade(true, 0);		\slowly fade to black
			SetVid($03);		\restore normal text display
			exit;
			];
		  $08:	Undo;			\Backspace = undo previous moves
		  $48, $4B, $4D, $50: DoMove	\arrow key scan codes
		other	[];			\ignore illegal keystrokes

		DrawFloorPlan;

		if Solved then
			[PlaySoundFile("SokoSolved.wav");
			Key:= ChIn(1);	\wait for key, read and discard char
			OpenI(1);	\discard possible scan code
			quit;
			];
		];			\move loop

	Fade(true, 0);			\slowly fade to black
	DelayUS(500_000);		\pause 1/2 second
	Level:= Level+1;		\next level
	];				\level loop
];	\Main
